from setuptools_scm import get_version

extensions = ["sphinx.ext.autodoc", "sphinx.ext.viewcode"]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "pthelma"
copyright = "2024, IRMASYS"
author = "Antonis Christofides"
version = get_version(root="..")
release = version
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
html_theme = "alabaster"
html_static_path = ["_static"]
htmlhelp_basename = "pthelma"
latex_elements = {}
latex_documents = [
    (
        master_doc,
        "pthelma.tex",
        "pthelma Documentation",
        "Antonis Christofides",
        "manual",
    )
]
texinfo_documents = [
    (
        master_doc,
        "pthelma",
        "pthelma Documentation",
        author,
        "pthelma",
        "Utilities for hydrological and meteorological time series processing",
        "Miscellaneous",
    )
]


def setup(app):
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )



from collections import namedtuple

from .calculation import Rocc

Threshold = namedtuple("Threshold", ["delta_t", "allowed_diff"])


def rocc(*, timeseries, thresholds, symmetric=False, flag="TEMPORAL"):
    return Rocc(timeseries, thresholds, symmetric, flag).execute()



import datetime as dt

import pandas as pd

from enhydris_api_client import EnhydrisApiClient
from htimeseries import HTimeseries


class TimeseriesCache(object):
    def __init__(self, timeseries_group):
        self.timeseries_group = timeseries_group

    def update(self):
        for item in self.timeseries_group:
            self.base_url = item["base_url"]
            if self.base_url[-1] != "/":
                self.base_url += "/"
            self.station_id = item["station_id"]
            self.timeseries_group_id = item["timeseries_group_id"]
            self.timeseries_id = item["timeseries_id"]
            self.auth_token = item["auth_token"]
            self.filename = item["file"]
            self._update_for_one_timeseries()

    def _update_for_one_timeseries(self):
        cached_ts = self._read_timeseries_from_cache_file()
        end_date = self._get_timeseries_end_date(cached_ts)
        start_date = end_date + dt.timedelta(minutes=1)
        new_ts = self._append_newer_timeseries(start_date, cached_ts)
        with open(self.filename, "w", encoding="utf-8") as f:
            new_ts.write(f, format=HTimeseries.FILE)

    def _read_timeseries_from_cache_file(self):
        try:
            with open(self.filename, newline="\n") as f:
                return HTimeseries(f)
        except (FileNotFoundError, ValueError):
            # If file is corrupted or nonexistent, continue with empty time series
            return HTimeseries()

    def _get_timeseries_end_date(self, timeseries):
        try:
            end_date = timeseries.data.index[-1]
        except IndexError:
            # Timeseries is totally empty; no start and end date
            end_date = dt.datetime(1, 1, 1, 0, 0, tzinfo=dt.timezone.utc)
        return end_date

    def _append_newer_timeseries(self, start_date, old_ts):
        with EnhydrisApiClient(self.base_url, token=self.auth_token) as api_client:
            ts = api_client.read_tsdata(
                self.station_id,
                self.timeseries_group_id,
                self.timeseries_id,
                start_date=start_date,
            )
            new_data = ts.data

            # For appending to work properly, both time series need to have the same
            # tz.
            if len(old_ts.data):
                new_data.index = new_data.index.tz_convert(old_ts.data.index.tz)
            else:
                old_ts.data.index = old_ts.data.index.tz_convert(new_data.index.tz)

            ts.data = pd.concat(
                [old_ts.data, new_data], verify_integrity=True, sort=False
            )
        return ts



# cython: language_level=3, linetrace=True
# distutils: define_macros=CYTHON_TRACE=1

import datetime as dt
cimport numpy as np
import numpy as np
import pandas as pd
from libc.math cimport isnan
from pandas.tseries.frequencies import to_offset

from htimeseries import HTimeseries

from .haggregate import RegularizationMode as RM


class RegularizeError(Exception):
    pass


def regularize(ts, new_date_flag="DATEINSERT", mode=RM.INTERVAL):
    # Sanity checks
    if not hasattr(ts, "time_step"):
        raise RegularizeError("The source time series does not specify a time step")
    try:
        pd.to_timedelta(to_offset(ts.time_step))
    except ValueError:
        raise RegularizeError(
            "The time step is malformed or is specified in months. Only time steps "
            "specified in minutes, hours or days are supported."
        )

    # Set metadata of result
    result = HTimeseries()
    attrs = (
        "unit",
        "timezone",
        "time_step",
        "interval_type",
        "variable",
        "precision",
        "location",
    )
    for attr in attrs:
        setattr(result, attr, getattr(ts, attr, None))
    if hasattr(ts, "title"):
        result.title = "Regularized " + ts.title
    if hasattr(ts, "comment"):
        result.comment = (
            "Created by regularizing step of timeseries that had this comment:\n\n"
            + ts.comment
        )

    # Return immediately if empty
    if len(ts.data) == 0:
        return result

    # Determine first and last timestamps
    step = pd.Timedelta(ts.time_step)
    first_timestamp_of_result = ts.data.index[0].round(step)
    last_timestamp_of_result = ts.data.index[-1].round(step)

    # Transform all pandas information to plain numpy, which is way faster and is also
    # supported by numba and Cython
    max_flags_length = max(ts.data["flags"].str.len()) + 1 + len(new_date_flag)
    flags_dtype = "U" + str(max_flags_length)
    ts_index = ts.data.index.values.astype(long)
    ts_values = ts.data["value"].values
    ts_flags = ts.data["flags"].values.astype(flags_dtype)
    result_step = np.timedelta64(step).astype(int) * 1000
    result_index = pd.date_range(
        first_timestamp_of_result, last_timestamp_of_result, freq=ts.time_step
    ).values
    result_values = np.full(len(result_index), np.nan, dtype=object)
    result_flags = np.full(len(result_index), "", dtype=flags_dtype)

    # Do the job
    _perform_regularization(
        result_index,
        result_values,
        result_flags,
        ts_index,
        ts_values,
        ts_flags,
        result_step,
        new_date_flag,
        mode.value,
    )

    result.data = pd.DataFrame(
        index=result_index,
        columns=["value", "flags"],
        data=np.vstack((result_values, result_flags)).transpose(),
    ).tz_localize(dt.timezone.utc).tz_convert(first_timestamp_of_result.tz)
    return result


def _perform_regularization(
    np.ndarray result_index,
    np.ndarray result_values,
    np.ndarray result_flags,
    np.ndarray ts_index,
    np.ndarray ts_values,
    np.ndarray ts_flags,
    long result_step,
    str new_date_flag,
    int mode,
):
    cdef int i, previous_pos
    cdef long t

    previous_pos = 0
    for i in range(result_index.size):
        t = result_index[i]
        result_values[i], result_flags[i], previous_pos = _get_record(
            ts_index,
            ts_values,
            ts_flags,
            t,
            result_step,
            new_date_flag,
            previous_pos,
            mode,
        )


def _get_record(
    np.ndarray ts_index,
    np.ndarray ts_values,
    np.ndarray ts_flags,
    long t,
    long result_step,
    str new_date_flag,
    int previous_pos,
    int mode,
):
    cdef int i, found, count
    cdef int nearest_i = -1
    cdef int INTERVAL = RM.INTERVAL.value
    cdef int INSTANTANEOUS = RM.INSTANTANEOUS.value

    # Return the source record if it already exists
    found = False
    for i in range(previous_pos, ts_index.size):
        if ts_index[i] == t and (mode == INTERVAL or not isnan(ts_values[i])):
            found = True
            break
        if ts_index[i] > t:
            break
    if found:
        return ts_values[i], ts_flags[i], i

    # Otherwise get the nearby record, if it exists
    start = t - result_step / 2
    end = t + result_step / 2
    count = 0
    for i in range(previous_pos, ts_index.size):
        ti = ts_index[i]
        if ti >= start and ti < end and (mode == INTERVAL or not isnan(ts_values[i])):
            count += 1
            nearest_i = _get_nearest(nearest_i, i, ts_index, ts_values, t, mode)
        if ts_index[i] >= end:
            i -= 1
            break
    if count < 1 or (count > 1 and mode == INTERVAL):
        return np.nan, "", i
    value = ts_values[nearest_i]
    flags = ts_flags[nearest_i]
    if flags:
        flags += " "
    flags += new_date_flag
    return value, flags, i + 1


def _get_nearest(
    int previous_nearest_i,
    int current_i,
    np.ndarray ts_index,
    np.ndarray ts_values,
    long t,
    int mode,
):
    if mode == RM.INTERVAL.value:
        # In that case it doesn't really matter which is the nearest, so long as it's
        # only one (which is checked elsewhere), so we return immediately.
        return current_i
    if previous_nearest_i < 0:
        return current_i
    current_distance = abs(t - ts_index[current_i])
    previous_distance = abs(t - ts_index[previous_nearest_i])
    if current_distance < previous_distance:
        return current_i
    else:
        return previous_nearest_i



import datetime as dt
import textwrap
from copy import copy
from io import StringIO
from unittest import mock

import pandas as pd
import requests

from htimeseries import HTimeseries

test_timeseries_csv = textwrap.dedent(
    """\
    2014-01-01 08:00,11.0,
    2014-01-02 08:00,12.0,
    2014-01-03 08:00,13.0,
    2014-01-04 08:00,14.0,
    2014-01-05 08:00,15.0,
    """
)
test_timeseries_htimeseries = HTimeseries(
    StringIO(test_timeseries_csv), default_tzinfo=dt.timezone(dt.timedelta(hours=2))
)
test_timeseries_csv_top = "".join(test_timeseries_csv.splitlines(keepends=True)[:-1])
test_timeseries_csv_bottom = test_timeseries_csv.splitlines(keepends=True)[-1]


def mock_session(**kwargs):
    """Mock requests.Session.

    Returns
        @mock.patch("requests.Session", modified_kwargs)

    However, it first tampers with kwargs in order to achieve the following:
    - It adds a leading "return_value." to the kwargs; so you don't need to specify,
      for example, "return_value.get.return_value", you just specify "get.return_value".
    - If kwargs doesn't contain "get.return_value.status_code", it adds
      a return code of 200. Likewise for post, put and patch. For delete it's 204.
    - If "get.return_value.status_code" is not between 200 and 399,
      then raise_for_status() will raise HTTPError. Likewise for the other methods.
    """
    for method in ("get", "post", "put", "patch", "delete"):
        default_value = 204 if method == "delete" else 200
        c = kwargs.setdefault(method + ".return_value.status_code", default_value)
        if c < 200 or c >= 400:
            method_side_effect = method + ".return_value.raise_for_status.side_effect"
            kwargs[method_side_effect] = requests.HTTPError
    for old_key in list(kwargs.keys()):
        kwargs["return_value." + old_key] = kwargs.pop(old_key)
    return mock.patch("requests.Session", **kwargs)


class AssertFrameEqualMixin:
    def assert_frame_equal(self, actual, expected):
        self.assertEqual(
            actual.index.tz.utcoffset(None), expected.index.tz.utcoffset(None)
        )
        pd.testing.assert_frame_equal(actual, expected, check_index_type=False)

    def assert_frame_loosely_equal(self, actual, expected):
        actual = copy(actual)
        expected = copy(expected)
        actual.index = actual.index.tz_convert("UTC")
        expected.index = expected.index.tz_convert("UTC")
        pd.testing.assert_frame_equal(actual, expected)



from unittest import TestCase

import requests

from enhydris_api_client import EnhydrisApiClient

from . import mock_session


class ListTimeseriesTestCase(TestCase):
    @mock_session(
        **{"get.return_value.json.return_value": {"results": [{"hello": "world"}]}}
    )
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.data = self.client.list_timeseries(41, 42)

    def test_makes_request(self):
        self.mock_requests_session.return_value.get.assert_called_once_with(
            "https://mydomain.com/api/stations/41/timeseriesgroups/42/timeseries/"
        )

    def test_returns_data(self):
        self.assertEqual(self.data, [{"hello": "world"}])


class GetTimeseriesTestCase(TestCase):
    @mock_session(**{"get.return_value.json.return_value": {"hello": "world"}})
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.data = self.client.get_timeseries(41, 42, 43)

    def test_makes_request(self):
        self.mock_requests_session.return_value.get.assert_called_once_with(
            "https://mydomain.com/api/stations/41/timeseriesgroups/42/timeseries/43/"
        )

    def test_returns_data(self):
        self.assertEqual(self.data, {"hello": "world"})


class GetStationOrTimeseriesErrorTestCase(TestCase):
    @mock_session(**{"get.return_value.status_code": 404})
    def setUp(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")

    def test_raises_exception_on_get_station_error(self):
        with self.assertRaises(requests.HTTPError):
            self.data = self.client.get_station(42)

    def test_raises_exception_on_get_timeseries_error(self):
        with self.assertRaises(requests.HTTPError):
            self.data = self.client.get_timeseries(41, 42, 43)


class PostTimeseriesTestCase(TestCase):
    @mock_session(**{"post.return_value.json.return_value": {"id": 43}})
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.data = self.client.post_timeseries(41, 42, data={"location": "Syria"})

    def test_makes_request(self):
        self.mock_requests_session.return_value.post.assert_called_once_with(
            "https://mydomain.com/api/stations/41/timeseriesgroups/42/timeseries/",
            data={"location": "Syria"},
        )

    def test_returns_id(self):
        self.assertEqual(self.data, 43)


class FailedPostTimeseriesTestCase(TestCase):
    @mock_session(**{"post.return_value.status_code": 404})
    def setUp(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")

    def test_raises_exception_on_error(self):
        with self.assertRaises(requests.HTTPError):
            self.client.post_timeseries(41, 42, data={"location": "Syria"})


class DeleteTimeseriesTestCase(TestCase):
    @mock_session()
    def test_makes_request(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.client.delete_timeseries(41, 42, 43)
        mock_requests_session.return_value.delete.assert_called_once_with(
            "https://mydomain.com/api/stations/41/timeseriesgroups/42/timeseries/43/"
        )

    @mock_session(**{"delete.return_value.status_code": 404})
    def test_raises_exception_on_error(self, mock_requests_delete):
        self.client = EnhydrisApiClient("https://mydomain.com")
        with self.assertRaises(requests.HTTPError):
            self.client.delete_timeseries(41, 42, 43)



import configparser
import datetime as dt
import logging
import os
import sys
import traceback

import click

from enhydris_cache import TimeseriesCache
from pthelma._version import __version__


class WrongValueError(configparser.Error):
    pass


class App:
    def __init__(self, configfilename):
        self.configfilename = configfilename

    def run(self):
        self.config = AppConfig(self.configfilename)
        self.config.read()
        self._setup_logger()
        self._execute_with_error_handling()

    def _execute_with_error_handling(self):
        self.logger.info("Starting enhydris-cache, " + dt.datetime.today().isoformat())
        try:
            self._execute()
        except Exception as e:
            self.logger.error(str(e))
            self.logger.debug(traceback.format_exc())
            self.logger.info(
                "enhydris-cache terminated with error, "
                + dt.datetime.today().isoformat()
            )
            raise click.ClickException(str(e))
        else:
            self.logger.info(
                "Finished enhydris-cache, " + dt.datetime.today().isoformat()
            )

    def _setup_logger(self):
        self.logger = logging.getLogger("spatialize")
        self._set_logger_handler()
        self.logger.setLevel(self.config.loglevel.upper())

    def _set_logger_handler(self):
        if getattr(self.config, "logfile", None):
            self.logger.addHandler(logging.FileHandler(self.config.logfile))
        else:
            self.logger.addHandler(logging.StreamHandler())

    def _execute(self):
        os.chdir(self.config.cache_dir)
        self.cache = TimeseriesCache(self.config.timeseries_group)
        self.cache.update()


class AppConfig:
    config_file_general_options = {
        "logfile": {"fallback": ""},
        "loglevel": {"fallback": "warning"},
        "cache_dir": {"fallback": os.getcwd()},
    }
    config_file_timeseries_options = {
        "base_url": {},
        "station_id": {},
        "timeseries_id": {},
        "timeseries_group_id": {},
        "auth_token": {"fallback": None},
        "file": {},
    }

    def __init__(self, configfilename):
        self.configfilename = configfilename

    def read(self):
        try:
            self._parse_config()
        except (OSError, configparser.Error) as e:
            sys.stderr.write(str(e))
            raise click.ClickException(str(e))

    def _parse_config(self):
        self._read_config_file()
        self._parse_general_section()
        self._parse_timeseries_sections()

    def _read_config_file(self):
        self.config = configparser.ConfigParser(interpolation=None)
        with open(self.configfilename) as f:
            self.config.read_file(f)

    def _parse_general_section(self):
        options = {
            opt: self.config.get("General", opt, **kwargs)
            for opt, kwargs in self.config_file_general_options.items()
        }
        for key, value in options.items():
            setattr(self, key, value)
        self._parse_log_level()

    def _parse_log_level(self):
        log_levels = ("ERROR", "WARNING", "INFO", "DEBUG")
        self.loglevel = self.loglevel.upper()
        if self.loglevel not in log_levels:
            raise WrongValueError("loglevel must be one of " + ", ".join(log_levels))

    def _parse_timeseries_sections(self):
        self.timeseries_group = []
        for section in self.config:
            if section in ("General", "DEFAULT"):
                continue
            item = self._read_section(section)
            self.timeseries_group.append(item)

    def _read_section(self, section):
        options = {
            opt: self.config.get(section, opt, **kwargs)
            for opt, kwargs in self.config_file_timeseries_options.items()
        }
        options["name"] = section
        try:
            options["station_id"] = int(options["station_id"])
            options["timeseries_group_id"] = int(options["timeseries_group_id"])
            options["timeseries_id"] = int(options["timeseries_id"])
        except ValueError:
            raise WrongValueError(
                '"{}" or "{}" or "{}" is not a valid integer'.format(
                    options["station_id"],
                    options["timeseries_group_id"],
                    options["timeseries_id"],
                )
            )
        return options


@click.command()
@click.argument("configfile")
@click.version_option(
    version=__version__, message="%(prog)s from pthelma v.%(version)s"
)
def main(configfile):
    """Spatial integration"""

    app = App(configfile)
    app.run()



import numpy as np
from osgeo import gdal, osr


def setup_test_raster(filename, value, timestamp=None, srid=4326, unit=None):
    """Save value, which is an np array, to a GeoTIFF file."""
    nodata = 1e8
    value[np.isnan(value)] = nodata
    f = gdal.GetDriverByName("GTiff").Create(filename, 3, 3, 1, gdal.GDT_Float32)
    try:
        if timestamp:
            f.SetMetadataItem("TIMESTAMP", timestamp.isoformat())
        if unit:
            f.SetMetadataItem("UNIT", unit)
        if srid == 4326:
            f.SetGeoTransform((22.0, 0.01, 0, 38.0, 0, -0.01))
        elif srid == 2100:
            f.SetGeoTransform((320000, 1000, 0, 4210000, 0, -1000))
        sr = osr.SpatialReference()
        sr.ImportFromEPSG(srid)
        if int(gdal.__version__.split(".")[0]) > 2:
            sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        f.SetProjection(sr.ExportToWkt())
        f.GetRasterBand(1).SetNoDataValue(nodata)
        f.GetRasterBand(1).WriteArray(value)
    finally:
        f = None



import datetime as dt
import os
import struct
from glob import glob
from math import isnan

import iso8601
import numpy as np
from affine import Affine
from django.contrib.gis.gdal import CoordTransform, SpatialReference
from django.contrib.gis.gdal.error import GDALException
from django.contrib.gis.geos import Point as GeoDjangoPoint
from osgeo import gdal, ogr, osr

from htimeseries import HTimeseries

gdal.UseExceptions()

NODATAVALUE = -(2.0**127)


def coordinates2point(x, y, srid=4326):
    point = ogr.Geometry(ogr.wkbPoint)
    sr = osr.SpatialReference()
    sr.ImportFromEPSG(srid)
    if int(gdal.__version__.split(".")[0]) > 2:
        sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    point.AssignSpatialReference(sr)
    point.AddPoint(x, y)
    return point


def idw(point, data_layer, alpha=1):
    data_layer.ResetReading()
    features = [f for f in data_layer if not isnan(f.GetField("value"))]
    distances = np.array([point.Distance(f.GetGeometryRef()) for f in features])
    values = np.array([f.GetField("value") for f in features])
    matches_station_exactly = abs(distances) < 1e-3
    if matches_station_exactly.any():
        invdistances = np.where(matches_station_exactly, 1, 0)
    else:
        invdistances = distances ** (-alpha)
    weights = invdistances / invdistances.sum()
    return (weights * values).sum()


def integrate(dataset, data_layer, target_band, funct, kwargs={}):
    mask = dataset.GetRasterBand(1).ReadAsArray() != 0

    # Create an array with the x co-ordinate of each grid point, and
    # one with the y co-ordinate of each grid point
    height, width = mask.shape
    x_left, x_step, d1, y_top, d2, y_step = dataset.GetGeoTransform()
    xcoords = np.arange(x_left + x_step / 2.0, x_left + x_step * width, x_step)
    ycoords = np.arange(y_top + y_step / 2.0, y_top + y_step * height, y_step)
    xarray, yarray = np.meshgrid(xcoords, ycoords)

    # Create a ufunc that makes the interpolation given the above arrays
    def interpolate_one_point(x, y, mask):
        if not mask:
            return np.nan
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(x, y)
        return funct(point, data_layer, **kwargs)

    interpolate = np.vectorize(interpolate_one_point, otypes=[np.float32])

    # Make the calculation
    result = interpolate(xarray, yarray, mask)
    result[np.isnan(result)] = NODATAVALUE
    target_band.SetNoDataValue(NODATAVALUE)
    target_band.WriteArray(result)


def create_ogr_layer_from_timeseries(filenames, epsg, data_source):
    # Prepare the co-ordinate transformation from WGS84 to epsg
    source_sr = osr.SpatialReference()
    source_sr.ImportFromEPSG(4326)
    target_sr = osr.SpatialReference()
    target_sr.ImportFromEPSG(epsg)
    if int(gdal.__version__.split(".")[0]) > 2:
        source_sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        target_sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    transform = osr.CoordinateTransformation(source_sr, target_sr)

    layer = data_source.CreateLayer("stations", target_sr)
    layer.CreateField(ogr.FieldDefn("filename", ogr.OFTString))
    for filename in filenames:
        with open(filename, newline="\n") as f:
            # The default_tzinfo doesn't matter because we don't care about the data,
            # we only use the location.
            ts = HTimeseries(f, default_tzinfo=dt.timezone.utc)
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(ts.location["abscissa"], ts.location["ordinate"])
        point.Transform(transform)
        f = ogr.Feature(layer.GetLayerDefn())
        f.SetGeometry(point)
        f.SetField("filename", filename)
        layer.CreateFeature(f)
    return layer


def _needs_calculation(output_filename, date, stations_layer):
    """
    Used by h_integrate to check whether the output file needs to be calculated
    or not. It does not need to be calculated if it already exists and has been
    calculated from all available data.
    """
    # Return immediately if output file does not exist
    if not os.path.exists(output_filename):
        return True

    # Get list of files which were used to calculate the output file
    fp = gdal.Open(output_filename)
    try:
        actual_input_files = fp.GetMetadataItem("INPUT_FILES")
        if actual_input_files is None:
            raise IOError(
                "{} does not contain the metadata item INPUT_FILES".format(
                    output_filename
                )
            )
    finally:
        fp = None  # Close file
    actual_input_files = set(actual_input_files.split("\n"))

    # Get list of files available for calculating the output file
    stations_layer.ResetReading()
    available_input_files = set(
        [
            station.GetField("filename")
            for station in stations_layer
            if os.path.exists(station.GetField("filename"))
        ]
    )

    # Which of these files have not been used?
    unused_files = available_input_files - actual_input_files

    # For each one of these files, check whether it has newly available data.
    # Upon finding one that does, the verdict is made: return True
    for filename in unused_files:
        with open(filename, newline="\n") as f:
            t = HTimeseries(f)
        try:
            value = t.data.loc[date, "value"]
            if not isnan(value):
                return True
        except KeyError:
            continue

    # We were unable to find data that had not already been used
    return False


def h_integrate(
    mask, stations_layer, date, output_filename_prefix, date_fmt, funct, kwargs
):
    date_fmt_for_filename = date.strftime(date_fmt).replace(" ", "-").replace(":", "-")
    output_filename = "{}-{}.tif".format(
        output_filename_prefix, date.strftime(date_fmt_for_filename)
    )
    if not _needs_calculation(output_filename, date, stations_layer):
        return

    # Read the time series values and add the 'value' attribute to
    # stations_layer. Also determine the unit of measurement.
    stations_layer.CreateField(ogr.FieldDefn("value", ogr.OFTReal))
    input_files = []
    unit_of_measurement = None
    stations_layer.ResetReading()
    for station in stations_layer:
        filename = station.GetField("filename")
        with open(filename, newline="\n") as f:
            t = HTimeseries(f)
        if unit_of_measurement is None and hasattr(t, "unit"):
            unit_of_measurement = t.unit
        try:
            value = t.data.loc[date, "value"]
        except KeyError:
            value = np.nan
        station.SetField("value", value)
        if not isnan(value):
            input_files.append(filename)
        stations_layer.SetFeature(station)
    if not input_files:
        return

    # Create destination data source
    output = gdal.GetDriverByName("GTiff").Create(
        output_filename, mask.RasterXSize, mask.RasterYSize, 1, gdal.GDT_Float32
    )
    output.SetMetadataItem("TIMESTAMP", date.strftime(date_fmt))
    output.SetMetadataItem("INPUT_FILES", "\n".join(input_files))
    output.SetMetadataItem("UNIT", unit_of_measurement)

    try:
        # Set geotransform and projection in the output data source
        output.SetGeoTransform(mask.GetGeoTransform())
        output.SetProjection(mask.GetProjection())

        # Do the integration
        integrate(mask, stations_layer, output.GetRasterBand(1), funct, kwargs)
    finally:
        # Close the dataset
        output = None


class PassepartoutPoint:
    """Uniform interface for GeoDjango Point and OGR Point."""

    def __init__(self, point):
        self.point = point

    def transform_to(self, target_srs_wkt):
        point = self.clone(self.point)
        if isinstance(self.point, GeoDjangoPoint):
            source_srs = point.srs or SpatialReference(4326)
            try:
                source_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
            except AttributeError:
                pass
            ct = CoordTransform(source_srs, SpatialReference(target_srs_wkt))
            point.transform(ct)
            return PassepartoutPoint(point)
        else:
            point_sr = point.GetSpatialReference()
            raster_sr = osr.SpatialReference()
            raster_sr.ImportFromWkt(target_srs_wkt)
            if int(gdal.__version__.split(".")[0]) > 2:
                point_sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
                raster_sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
            transform = osr.CoordinateTransformation(point_sr, raster_sr)
            point.Transform(transform)
            return PassepartoutPoint(point)

    def clone(self, original_point):
        if isinstance(original_point, GeoDjangoPoint):
            return GeoDjangoPoint(
                original_point.x, original_point.y, original_point.srid
            )
        else:
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(original_point.GetX(), original_point.GetY())
            point.AssignSpatialReference(original_point.GetSpatialReference())
            return point

    @property
    def x(self):
        try:
            return self.point.x
        except AttributeError:
            return self.point.GetX()

    @property
    def y(self):
        try:
            return self.point.y
        except AttributeError:
            return self.point.GetY()


def extract_point_from_raster(point, data_source, band_number=1):
    """Return floating-point value that corresponds to given point."""
    pppoint = PassepartoutPoint(point)

    # Convert point co-ordinates so that they are in same projection as raster
    try:
        target_srs_wkt = data_source.GetProjection()
    except AttributeError:
        target_srs_wkt = data_source.srs.wkt
    try:
        pppoint = pppoint.transform_to(target_srs_wkt)
    except GDALException:
        raise RuntimeError("Couldn't convert point to raster's CRS")
    infinities = (float("inf"), float("-inf"))
    if pppoint.x in infinities or pppoint.y in infinities:
        raise RuntimeError("Couldn't convert point to raster's CRS")

    # Convert geographic co-ordinates to pixel co-ordinates
    try:
        forward_transform = Affine.from_gdal(*data_source.GetGeoTransform())
    except AttributeError:
        forward_transform = Affine.from_gdal(*data_source.geotransform)
    reverse_transform = ~forward_transform
    px, py = reverse_transform * (pppoint.x, pppoint.y)
    px, py = int(px), int(py)

    # Extract pixel value
    try:
        band = data_source.GetRasterBand(band_number)
    except AttributeError:
        band = data_source.bands[band_number - 1]
    try:
        structval = band.ReadRaster(px, py, 1, 1, buf_type=gdal.GDT_Float32)
    except AttributeError:
        structval = band.data(offset=(px, py), size=(1, 1))
    result = struct.unpack("f", structval)[0]
    try:
        nodata_value = band.GetNoDataValue()
    except AttributeError:
        nodata_value = band.nodata_value
    if result == nodata_value:
        result = float("nan")
    return result


class PointTimeseries:
    def __init__(self, point, **kwargs):
        self.point = point
        filenames = kwargs.pop("filenames", None)
        self.prefix = kwargs.pop("prefix", None)
        assert filenames is None or self.prefix is None
        assert filenames is not None or self.prefix is not None
        self.date_fmt = kwargs.pop("date_fmt", None)
        self.start_date = kwargs.pop("start_date", None)
        self.end_date = kwargs.pop("end_date", None)
        self.default_time = kwargs.pop(
            "default_time", dt.time(0, 0, tzinfo=dt.timezone.utc)
        )
        if self.default_time.tzinfo is None:
            raise TypeError("default_time must be aware")
        if self.start_date and self.start_date.tzinfo is None:
            self.start_date = self.start_date.replace(tzinfo=self.default_time.tzinfo)
        if self.end_date and self.end_date.tzinfo is None:
            self.end_date = self.end_date.replace(tzinfo=self.default_time.tzinfo)
        self.filenames = self._get_filenames(filenames)

    def _get_filenames(self, filenames):
        if self.prefix is None:
            return filenames
        filenames = glob(self.prefix + "-*.tif")
        self.filename_format = FilenameWithDateFormat(
            self.prefix, date_fmt=self.date_fmt, tzinfo=self.default_time.tzinfo
        )
        result = []
        for filename in filenames:
            date = self.filename_format.get_date(filename)
            is_after_start_date = (self.start_date is None) or (date >= self.start_date)
            is_before_end_date = (self.end_date is None) or (date <= self.end_date)
            if is_after_start_date and is_before_end_date:
                result.append(filename)
        return result

    def get(self):
        result = HTimeseries(default_tzinfo=self.default_time.tzinfo)
        for filename in self.filenames:
            f = gdal.Open(filename)
            try:
                timestamp = self._get_timestamp(f)
                self._get_unit_of_measurement(f, result)
                value = extract_point_from_raster(self.point, f)
                result.data.loc[timestamp, "value"] = value
                result.data.loc[timestamp, "flags"] = ""
            finally:
                f = None
        result.data = result.data.sort_index()
        return result

    def _get_timestamp(self, f):
        isostring = f.GetMetadata()["TIMESTAMP"]
        timestamp = iso8601.parse_date(
            isostring, default_timezone=self.default_time.tzinfo
        )
        if len(isostring) <= 10:
            timestamp = dt.datetime.combine(timestamp.date(), self.default_time)
        return timestamp

    def _get_unit_of_measurement(self, f, ahtimeseries):
        if hasattr(ahtimeseries, "unit"):
            return
        unit = f.GetMetadataItem("UNIT")
        if unit is not None:
            ahtimeseries.unit = unit

    def get_cached(self, dest, force=False, version=4):
        assert self.prefix
        ts = self._get_saved_timeseries_if_updated_else_none(dest, force)
        if ts is None:
            ts = self.get()
            with open(dest, "w", newline="") as f:
                ts.write(f, format=HTimeseries.FILE, version=version)
        return ts

    def _get_saved_timeseries_if_updated_else_none(self, dest, force):
        if force or not os.path.exists(dest):
            return None
        else:
            return self._get_timeseries_if_file_is_up_to_date_else_none(dest)

    def _get_timeseries_if_file_is_up_to_date_else_none(self, dest):
        with open(dest, "r", newline="") as f:
            ts = HTimeseries(f, default_tzinfo=self.default_time.tzinfo)
        for filename in self.filenames:
            if not self.filename_format.get_date(filename) in ts.data.index:
                return None
        return ts


class FilenameWithDateFormat:
    def __init__(self, prefix, *, date_fmt=None, tzinfo):
        self.prefix = prefix
        self.date_fmt = date_fmt
        self.tzinfo = tzinfo

    def get_date(self, filename):
        datestr = self._extract_datestr(filename)
        self._ensure_we_have_date_fmt(datestr)
        return dt.datetime.strptime(datestr, self.date_fmt).replace(tzinfo=self.tzinfo)

    def _ensure_we_have_date_fmt(self, datestr):
        if self.date_fmt is not None:
            pass
        elif datestr.count("-") == 4:
            self.date_fmt = "%Y-%m-%d-%H-%M"
        elif datestr.count("-") == 2:
            self.date_fmt = "%Y-%m-%d"
        else:
            raise ValueError("Invalid date " + datestr)

    def _extract_datestr(self, filename):
        assert filename.startswith(self.prefix + "-")
        assert filename.endswith(".tif")
        startpos = len(self.prefix) + 1
        return filename[startpos:-4]



import datetime as dt
import json
import os
import textwrap
from io import StringIO
from unittest import TestCase, skipUnless

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

import pandas as pd
import requests

from enhydris_api_client import EnhydrisApiClient
from htimeseries import HTimeseries

from . import AssertFrameEqualMixin, test_timeseries_htimeseries

UTC_PLUS_2 = dt.timezone(dt.timedelta(hours=2))


@skipUnless(
    os.getenv("PTHELMA_TEST_ENHYDRIS_SERVER"), "Set PTHELMA_TEST_ENHYDRIS_SERVER"
)
class EndToEndTestCase(AssertFrameEqualMixin, TestCase):
    """End-to-end test against a real Enhydris instance.
    To execute this test, specify the PTHELMA_TEST_ENHYDRIS_SERVER environment variable
    like this:
        PTHELMA_TEST_ENHYDRIS_SERVER='
            {"base_url": "http://localhost:8001",
             "token": "topsecrettokenkey",
             "owner_id": 9,
             "variable_id": 22,
             "unit_of_measurement_id": 18,
             "station_id": 1403,
             "timeseries_group_id": 513
             }
        '
    This should point to an Enhydris server. Avoid using a production database for
    that; the testing functionality will write objects to the database. Although
    things are normally cleaned up (created objects will be deleted), id serial
    numbers will be affected and things might not be cleaned up if there is an error.

    It's possible that not all items of the above PTHELMA_TEST_ENHYDRIS_SERVER example
    are being used in the test case, but they might be used by other test cases, such
    as in enhydris_cache.
    """

    def setUp(self):
        v = json.loads(os.getenv("PTHELMA_TEST_ENHYDRIS_SERVER"))
        self.token = v["token"]
        self.client = EnhydrisApiClient(v["base_url"], token=self.token)
        self.client.__enter__()
        self.owner_id = v["owner_id"]
        self.variable_id = v["variable_id"]
        self.unit_of_measurement_id = v["unit_of_measurement_id"]

    def tearDown(self):
        self.client.__exit__()

    def test_e2e(self):
        # Verify we're authenticated
        token = self.client.session.headers.get("Authorization")
        self.assertEqual(token, f"token {self.token}")

        # Create a test station
        tmp_station_id = self.client.post_station(
            {
                "name": "My station",
                "copyright_holder": "Joe User",
                "copyright_years": "2019",
                "geom": "POINT(20.94565 39.12102)",
                "original_srid": 4326,
                "owner": self.owner_id,
                "display_timezone": "Etc/GMT-1",
            }
        )

        # Get the test station
        station = self.client.get_station(tmp_station_id)
        self.assertEqual(station["id"], tmp_station_id)
        self.assertEqual(station["name"], "My station")

        # Patch station and verify
        self.client.patch_station(tmp_station_id, {"name": "New name"})
        station = self.client.get_station(tmp_station_id)
        self.assertEqual(station["name"], "New name")

        # Put station and verify
        self.client.put_station(
            tmp_station_id,
            {
                "name": "Newer name",
                "copyright_holder": "Joe User",
                "copyright_years": "2019",
                "geom": "POINT(20.94565 39.12102)",
                "original_srid": 4326,
                "owner": self.owner_id,
            },
        )
        station = self.client.get_station(tmp_station_id)
        self.assertEqual(station["name"], "Newer name")

        # Create a time series group and verify it was created
        self.timeseries_group_id = self.client.post_timeseries_group(
            tmp_station_id,
            data={
                "name": "My time series group",
                "gentity": tmp_station_id,
                "variable": self.variable_id,
                "unit_of_measurement": self.unit_of_measurement_id,
                "precision": 2,
            },
        )
        timeseries_group = self.client.get_timeseries_group(
            tmp_station_id, self.timeseries_group_id
        )
        self.assertEqual(timeseries_group["name"], "My time series group")

        # Create a time series and verify it was created
        self.timeseries_id = self.client.post_timeseries(
            tmp_station_id,
            self.timeseries_group_id,
            data={
                "type": "Regularized",
                "time_step": "10min",
                "timeseries_group": self.timeseries_group_id,
            },
        )
        timeseries = self.client.get_timeseries(
            tmp_station_id, self.timeseries_group_id, self.timeseries_id
        )
        self.assertEqual(timeseries["type"], "Regularized")

        # Post time series data
        self.client.post_tsdata(
            tmp_station_id,
            self.timeseries_group_id,
            self.timeseries_id,
            test_timeseries_htimeseries,
        )

        # Get the last date and check it
        date = self.client.get_ts_end_date(
            tmp_station_id,
            self.timeseries_group_id,
            self.timeseries_id,
        )
        self.assertEqual(date, dt.datetime(2014, 1, 5, 7, 0))

        # Get the last date in a different timezone from the default
        date = self.client.get_ts_end_date(
            tmp_station_id,
            self.timeseries_group_id,
            self.timeseries_id,
            timezone="Etc/GMT-5",
        )
        self.assertEqual(date, dt.datetime(2014, 1, 5, 11, 0))

        # Get all time series data and check it
        hts = self.client.read_tsdata(
            tmp_station_id, self.timeseries_group_id, self.timeseries_id
        )
        try:
            # Compatibility with older Python or pandas versions (such as Python 3.7
            # with pandas 0.23): comparison may fail if tzinfo, although practically the
            # same thing, is a different object
            if hts.data.index.tz.offset == dt.timedelta(0):
                hts.data.index = hts.data.index.tz_convert(dt.timezone.utc)
        except AttributeError:
            pass
        self.assert_frame_loosely_equal(hts.data, test_timeseries_htimeseries.data)

        # The other attributes should have been set too.
        self.assertTrue(hasattr(hts, "variable"))

        # Get part of the time series data and check it
        hts = self.client.read_tsdata(
            tmp_station_id,
            self.timeseries_group_id,
            self.timeseries_id,
            start_date=dt.datetime(2014, 1, 3, 8, 0, tzinfo=UTC_PLUS_2),
            end_date=dt.datetime(2014, 1, 4, 8, 0, tzinfo=UTC_PLUS_2),
            timezone="Etc/GMT-1",
        )
        expected_result = HTimeseries(
            StringIO(
                textwrap.dedent(
                    """\
                    2014-01-03 07:00,13.0,
                    2014-01-04 07:00,14.0,
                    """
                )
            ),
            default_tzinfo=ZoneInfo("Etc/GMT-1"),
        )
        try:
            # Compatibility with older Python or pandas versions (such as Python 3.7
            # with pandas 0.23): comparison may fail if tzinfo, although practically the
            # same thing, is a different object
            if hts.data.index.tz.offset == dt.timedelta(minutes=60):
                hts.data.index = hts.data.index.tz_convert(ZoneInfo("Etc/GMT-1"))
        except AttributeError:
            pass
        pd.testing.assert_frame_equal(hts.data, expected_result.data)

        # Delete the time series and verify
        self.client.delete_timeseries(
            tmp_station_id, self.timeseries_group_id, self.timeseries_id
        )
        with self.assertRaises(requests.HTTPError):
            self.client.get_timeseries(
                tmp_station_id, self.timeseries_group_id, self.timeseries_id
            )

        # Delete station
        self.client.delete_station(tmp_station_id)
        with self.assertRaises(requests.HTTPError):
            self.client.get_station(tmp_station_id)



import textwrap
from io import StringIO
from unittest import TestCase

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from htimeseries import HTimeseries
from rocc import Threshold, rocc


class RoccTestCase(TestCase):
    test_data = textwrap.dedent(
        """\
        2020-10-06 14:30,24.0,
        2020-10-06 14:40,25.0,
        2020-10-06 14:50,36.0,SOMEFLAG
        2020-10-06 15:01,51.0,
        2020-10-06 15:21,55.0,
        2020-10-06 15:31,65.0,
        2020-10-06 15:41,75.0,
        2020-10-06 15:51,70.0,
        """
    )

    def setUp(self):
        self.ahtimeseries = HTimeseries(
            StringIO(self.test_data), default_tzinfo=ZoneInfo("Etc/GMT-2")
        )
        self.ahtimeseries.precision = 1

    def _run_rocc(self, flag):
        self.return_value = rocc(
            timeseries=self.ahtimeseries,
            thresholds=(
                Threshold("10min", 10),
                Threshold("20min", 15),
                Threshold("h", 40),
            ),
            flag=flag,
        )

    def test_calculation(self):
        self._run_rocc(flag="TEMPORAL")
        result = StringIO()
        self.ahtimeseries.write(result)
        result = result.getvalue().replace("\r\n", "\n")
        self.assertEqual(
            result,
            textwrap.dedent(
                """\
                2020-10-06 14:30,24.0,
                2020-10-06 14:40,25.0,
                2020-10-06 14:50,36.0,SOMEFLAG TEMPORAL
                2020-10-06 15:01,51.0,
                2020-10-06 15:21,55.0,
                2020-10-06 15:31,65.0,
                2020-10-06 15:41,75.0,TEMPORAL
                2020-10-06 15:51,70.0,
                """
            ),
        )

    def test_return_value(self):
        self._run_rocc(flag="TEMPORAL")
        self.assertEqual(len(self.return_value), 2)
        self.assertEqual(
            self.return_value[0], "2020-10-06T14:50  +11.0 in 10min (> 10.0)"
        )
        self.assertEqual(
            self.return_value[1], "2020-10-06T15:41  +20.0 in 20min (> 15.0)"
        )

    def test_value_dtype(self):
        self._run_rocc(flag="TEMPORAL")
        expected_dtype = HTimeseries().data["value"].dtype
        self.assertEqual(self.ahtimeseries.data["value"].dtype, expected_dtype)

    def test_flags_dtype(self):
        self._run_rocc(flag="TEMPORAL")
        expected_dtype = HTimeseries().data["flags"].dtype
        self.assertEqual(self.ahtimeseries.data["flags"].dtype, expected_dtype)

    def test_empty_flag(self):
        self._run_rocc(flag=None)
        result = StringIO()
        self.ahtimeseries.write(result)
        result = result.getvalue().replace("\r\n", "\n")
        self.assertEqual(result, self.test_data)


class RoccSymmetricCase(TestCase):
    test_data = textwrap.dedent(
        """\
        2020-10-06 14:30,76.0,
        2020-10-06 14:40,75.0,SOMEFLAG
        2020-10-06 14:50,64.0,SOMEFLAG
        2020-10-06 15:01,49.0,
        2020-10-06 15:21,45.0,
        2020-10-06 15:31,35.0,
        2020-10-06 15:41,25.0,
        2020-10-06 15:51,30.0,
        """
    )

    def setUp(self):
        self.ahtimeseries = HTimeseries(
            StringIO(self.test_data), default_tzinfo=ZoneInfo("Etc/GMT-2")
        )
        self.ahtimeseries.precision = 1

    def test_without_symmetric(self):
        rocc(
            timeseries=self.ahtimeseries,
            thresholds=(
                Threshold("10min", 10),
                Threshold("20min", 15),
                Threshold("h", 40),
            ),
        )
        result = StringIO()
        self.ahtimeseries.write(result)
        result = result.getvalue().replace("\r\n", "\n")
        self.assertEqual(
            result,
            textwrap.dedent(
                """\
                2020-10-06 14:30,76.0,
                2020-10-06 14:40,75.0,SOMEFLAG
                2020-10-06 14:50,64.0,SOMEFLAG
                2020-10-06 15:01,49.0,
                2020-10-06 15:21,45.0,
                2020-10-06 15:31,35.0,
                2020-10-06 15:41,25.0,
                2020-10-06 15:51,30.0,
                """
            ),
        )

    def test_with_symmetric(self):
        rocc(
            timeseries=self.ahtimeseries,
            thresholds=(
                Threshold("10min", 10),
                Threshold("20min", 15),
                Threshold("h", 40),
            ),
            symmetric=True,
        )
        result = StringIO()
        self.ahtimeseries.write(result)
        result = result.getvalue().replace("\r\n", "\n")
        self.assertEqual(
            result,
            textwrap.dedent(
                """\
                2020-10-06 14:30,76.0,
                2020-10-06 14:40,75.0,SOMEFLAG
                2020-10-06 14:50,64.0,SOMEFLAG TEMPORAL
                2020-10-06 15:01,49.0,
                2020-10-06 15:21,45.0,
                2020-10-06 15:31,35.0,
                2020-10-06 15:41,25.0,TEMPORAL
                2020-10-06 15:51,30.0,
                """
            ),
        )

    def test_symmetric_return_value(self):
        return_value = rocc(
            timeseries=self.ahtimeseries,
            thresholds=(
                Threshold("10min", 10),
                Threshold("20min", 15),
                Threshold("h", 40),
            ),
            symmetric=True,
        )
        self.assertEqual(len(return_value), 2)
        self.assertEqual(return_value[0], "2020-10-06T14:50  -11.0 in 10min (< -10.0)")
        self.assertEqual(return_value[1], "2020-10-06T15:41  -20.0 in 20min (< -15.0)")


class RoccEmptyCase(TestCase):
    def test_with_empty(self):
        ahtimeseries = HTimeseries()
        rocc(timeseries=ahtimeseries, thresholds=[Threshold("10min", 10)])
        self.assertTrue(ahtimeseries.data.empty)



import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup

setup(
    ext_modules=cythonize(
        [
            Extension(
                "haggregate.regularize",
                sources=["src/haggregate/regularize.pyx"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                "rocc.calculation",
                sources=["src/rocc/calculation.pyx"],
                include_dirs=[numpy.get_include()],
            ),
        ]
    ),
)



from copy import copy
from io import StringIO
from urllib.parse import urljoin

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

import iso8601
import requests

from htimeseries import HTimeseries


class EnhydrisApiClient:
    def __init__(self, base_url, token=None):
        self.base_url = base_url
        self.token = token
        self.session = requests.Session()
        if token is not None:
            self.session.headers.update({"Authorization": f"token {self.token}"})

        # My understanding from requests' documentation is that when I make a post
        # request, it shouldn't be necessary to specify Content-Type:
        # application/x-www-form-urlencoded, and that requests adds the header
        # automatically. However, when running in Python 3, apparently requests does not
        # add the header (although it does convert the post data to
        # x-www-form-urlencoded format). This is why I'm specifying it explicitly.
        self.session.headers.update(
            {"Content-Type": "application/x-www-form-urlencoded"}
        )

    def __enter__(self):
        self.session.__enter__()
        return self

    def __exit__(self, *args):
        self.session.__exit__(*args)

    def check_response(self, expected_status_code=None):
        try:
            self._raise_HTTPError_on_error(expected_status_code=expected_status_code)
        except requests.HTTPError as e:
            if self.response.text:
                raise requests.HTTPError(
                    f"{str(e)}. Server response: {self.response.text}"
                )

    def _raise_HTTPError_on_error(self, expected_status_code):
        self._check_status_code_is_nonerror()
        self._check_status_code_is_the_one_expected(expected_status_code)

    def _check_status_code_is_nonerror(self):
        self.response.raise_for_status()

    def _check_status_code_is_the_one_expected(self, expected_status_code):
        if expected_status_code and self.response.status_code != expected_status_code:
            raise requests.HTTPError(
                f"Expected status code {expected_status_code}; "
                f"got {self.response.status_code} instead"
            )

    def get_token(self, username, password):
        if not username:
            return

        # Get a csrftoken
        login_url = urljoin(self.base_url, "api/auth/login/")
        data = f"username={username}&password={password}"
        self.response = self.session.post(login_url, data=data, allow_redirects=False)
        self.check_response()
        key = self.response.json()["key"]
        self.session.headers.update({"Authorization": f"token {key}"})
        return key

    def get_station(self, station_id):
        url = urljoin(self.base_url, f"api/stations/{station_id}/")
        self.response = self.session.get(url)
        self.check_response()
        return self.response.json()

    def post_station(self, data):
        self.response = self.session.post(
            urljoin(self.base_url, "api/stations/"), data=data
        )
        self.check_response()
        return self.response.json()["id"]

    def put_station(self, station_id, data):
        self.response = self.session.put(
            urljoin(self.base_url, f"api/stations/{station_id}/"), data=data
        )
        self.check_response()

    def patch_station(self, station_id, data):
        self.response = self.session.patch(
            urljoin(self.base_url, f"api/stations/{station_id}/"), data=data
        )
        self.check_response()

    def delete_station(self, station_id):
        url = urljoin(self.base_url, f"api/stations/{station_id}/")
        self.response = self.session.delete(url)
        self.check_response(expected_status_code=204)

    def get_timeseries_group(self, station_id, timeseries_group_id):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/",
        )
        self.response = self.session.get(url)
        self.check_response()
        return self.response.json()

    def post_timeseries_group(self, station_id, data):
        url = urljoin(self.base_url, f"api/stations/{station_id}/timeseriesgroups/")
        self.response = self.session.post(url, data=data)
        self.check_response()
        return self.response.json()["id"]

    def put_timeseries_group(self, station_id, timeseries_group_id, data):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/",
        )
        self.response = self.session.put(url, data=data)
        self.check_response()
        return self.response.json()["id"]

    def patch_timeseries_group(self, station_id, timeseries_group_id, data):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/",
        )
        self.response = self.session.patch(url, data=data)
        self.check_response()

    def delete_timeseries_group(self, station_id, timeseries_group_id):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/",
        )
        self.response = self.session.delete(url)
        self.check_response(expected_status_code=204)

    def list_timeseries(self, station_id, timeseries_group_id):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
            "timeseries/",
        )
        self.response = self.session.get(url)
        self.check_response()
        return self.response.json()["results"]

    def get_timeseries(self, station_id, timeseries_group_id, timeseries_id):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
            f"timeseries/{timeseries_id}/",
        )
        self.response = self.session.get(url)
        self.check_response()
        return self.response.json()

    def post_timeseries(self, station_id, timeseries_group_id, data):
        self.response = self.session.post(
            urljoin(
                self.base_url,
                f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
                "timeseries/",
            ),
            data=data,
        )
        self.check_response()
        return self.response.json()["id"]

    def delete_timeseries(self, station_id, timeseries_group_id, timeseries_id):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
            f"timeseries/{timeseries_id}/",
        )
        self.response = self.session.delete(url)
        self.check_response(expected_status_code=204)

    def read_tsdata(
        self,
        station_id,
        timeseries_group_id,
        timeseries_id,
        start_date=None,
        end_date=None,
        timezone=None,
    ):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
            f"timeseries/{timeseries_id}/data/",
        )
        params = {"fmt": "hts"}
        tzinfo = ZoneInfo(timezone) if timezone else None
        dates_are_aware = (start_date is None or start_date.tzinfo is not None) and (
            end_date is None or end_date.tzinfo is not None
        )
        if not dates_are_aware:
            raise ValueError("start_date and end_date must be aware")
        params["start_date"] = start_date and start_date.isoformat()
        params["end_date"] = end_date and end_date.isoformat()
        params["timezone"] = timezone
        self.response = self.session.get(url, params=params)
        self.check_response()
        if self.response.text:
            return HTimeseries(StringIO(self.response.text), default_tzinfo=tzinfo)
        else:
            return HTimeseries()

    def post_tsdata(self, station_id, timeseries_group_id, timeseries_id, ts):
        f = StringIO()
        data = copy(ts.data)
        try:
            data.index = data.index.tz_convert("UTC")
        except AttributeError:
            assert data.empty
        data.to_csv(f, header=False)
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
            f"timeseries/{timeseries_id}/data/",
        )
        self.response = self.session.post(
            url, data={"timeseries_records": f.getvalue(), "timezone": "UTC"}
        )
        self.check_response()
        return self.response.text

    def get_ts_end_date(
        self, station_id, timeseries_group_id, timeseries_id, timezone=None
    ):
        url = urljoin(
            self.base_url,
            f"api/stations/{station_id}/timeseriesgroups/{timeseries_group_id}/"
            f"timeseries/{timeseries_id}/bottom/",
        )
        self.response = self.session.get(url, params={"timezone": timezone})
        self.check_response()
        try:
            datestring = self.response.text.strip().split(",")[0]
            return iso8601.parse_date(datestring, default_timezone=None)
        except (IndexError, iso8601.ParseError):
            return None
